from typing import Dict, List, Tuple

import numpy as np
import slugify
from bson.objectid import ObjectId

from efemarai.fields.base_fields import BaseField, sdk_serialize


def chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i : i + n]


class AnnotationClass:
    """
    Represents an annotation that can be attached to any other field and can
    act as a label. When returned by a model output, either name or id can be
    specified. When creating a dataset, both are required.

    Args:
        name (str): Name of the class.
        id (int): Id of the class.
        confidence (Optional[float]): Confidence level. Prefer using the DataField confidence in general.
        color (Optional[str]): Color of the class in hex string. Used for visualization.
        text_color (Optional[str]): Color of the text of the annotation. Used for visualization.
        category (Optional[str]): Category of the class.
        slug (Optional[str]): Slugified name of the class that can be used for queries etc.

    Returns:
        :class: `efemarai.fields.annotation_fields.AnnotationClass`: An AnnotationClass object.

    Example:

    When your code needs to return a target that has a label, you should create one with
    the name _or_ id that your model returns.

    .. code-block:: python

        import efemarai as ef
        output = {"class_id": 1, "name": "person", "value": 0.77}
        label = ef.AnnotationClass(id=output["class_id"], name=output["name"])
        label2 = ef.AnnotationClass(id=output["class_id"])
        print(label, label2)

    If you are appending ground truth data to a dataset, you need to resolve all of the information.
    You can do that through the dataset object to obtain the annotation.

    .. code-block:: python

        label = dataset.get_annotation_class(id=output["class_id"])
        label = dataset.get_annotation_class(name=output["name"])
    """

    def __init__(
        self,
        name: str = None,
        id: int = None,
        confidence: float = None,
        color: str = None,
        text_color: str = None,
        category: str = None,
        slug: str = None,
    ):
        if name is None and id is None:
            raise AssertionError("'name' and 'id' cannot be None at the same time.")

        self.name = name
        self.id = int(id)
        self.confidence = confidence
        self.category = category
        self.color = color if color is not None else self.get_random_color()
        self.text_color = (
            text_color if text_color is not None else self.get_random_color()
        )
        self.slug = slug if slug is not None else self.get_slug()

    def get_slug(self):
        return slugify.slugify(self.name) if self.name is not None else None

    @staticmethod
    def get_random_color():
        import random

        def r():
            return random.randint(0, 255)

        return f"#{r():02X}{r():02X}{r():02X}"

    def _serialize(self):
        return sdk_serialize(self)

    def __repr__(self):
        res = f"{self.__module__}.{self.__class__.__name__}("
        res += f"\n  name={self.name}"
        res += f"\n  id={self.id}"
        res += f"\n  confidence={self.confidence}"
        res += f"\n  color={self.color}" if self.color else ""
        res += f"\n  text_color={self.text_color}" if self.text_color else ""
        res += f"\n  category={self.category}" if self.category else ""
        res += f"\n  slug={self.slug}"
        res += "\n)"
        return res


class Tag(BaseField):
    """
    Represents an annotation that can be used for classification.

    Args:
        label (AnnotationClass): An instance of AnnotationClass
        probabilities (np.ndarray): The softmax probabilities for a given prediction.
        confidence (Optional[float]): Confidence level of the field (autocalculated from probabilities).
        id (Optional[ObjectID]): Id of the field.
        description (Optional[str]): A description of the object or annotation information.
        ref_field (List[object]): A list of objects the current BaseField is refering to.
        key_name (Optional[str]): A string key, by which the field can be identified.
        user_attributes (Optional[Dict[str, object]]): Optional dictionary containing additional data regarding the field.

    Returns:
        :class: `efemarai.fields.annotation_fields.Tag`: A Tag object.

    Example:

    .. code-block:: python

        import efemarai as ef
        import numpy as np

        outputs = [0.1, 0.2, 0.15, 0.55]

        label = ef.AnnotationClass(id=np.argmax(outputs))
        tag = ef.Tag(label=label, probabilities=outputs)
        print(label, tag)
    """

    def __init__(
        self,
        label: AnnotationClass,
        probabilities: np.ndarray = None,
        confidence: float = None,
        id: ObjectId = None,
        description: str = None,
        ref_field: list = None,
        key_name: str = None,
        user_attributes: Dict[str, object] = None,
    ):
        probs = np.asarray(probabilities)
        # Get confidence from passed arg, or from probabilities if they exist
        confidence = (
            confidence
            if confidence is not None
            else float(
                (-1 * probs * np.log2(probs)).sum(axis=probs.ndim - 1)
            )  # Could we use multidim data?
            if probs.any()
            else None
        )
        super().__init__(
            confidence=confidence,
            id=id,
            description=description,
            ref_field=ref_field,
            key_name=key_name,
            user_attributes=user_attributes,
        )

        self.probabilities = probabilities
        self.label = label

    def __repr__(self):
        res = f"{self.__module__}.{self.__class__.__name__}("
        res += f"\n  label={self.label}"
        res += f"\n  probabilities={self.probabilities}"
        res += f"\n  confidence={self.confidence}"
        res += f"\n  description={self.description}"
        res += f"\n  ref_field={self.ref_field}"
        res += f"\n  key_name={self.key_name}"
        res += "\n)"
        return res


class Value(BaseField):
    """
    Represents an annotation that can be used for regression.

    Args:
        value (float): A float to represent the annotation value
        confidence (Optional[float]): An optional confidence associated with the value
        description (Optional[str]): A description of the object or annotation information.
        ref_field (List[object]): A list of objects the current BaseField is refering to.
        key_name (Optional[str]): A string key, by which the field can be identified.
        user_attributes (Optional[Dict[str, object]]): Dictionary containing additional data regarding the field.

    Returns:
        :class: `efemarai.fields.annotation_fields.Value`: A Value object.

    Example:

    .. code-block:: python

        import efemarai as ef
        import numpy as np

        output = 0.55

        value = ef.Value(value=output)
        print(value)
    """

    def __init__(
        self,
        value: float,
        confidence: float = None,
        id: ObjectId = None,
        description: str = None,
        ref_field: list = None,
        key_name: str = None,
        user_attributes: Dict[str, object] = None,
    ):
        super().__init__(
            confidence=confidence,
            id=id,
            description=description,
            ref_field=ref_field,
            key_name=key_name,
            user_attributes=user_attributes,
        )

        self.value = value

    def __repr__(self):
        res = f"{self.__module__}.{self.__class__.__name__}("
        res += f"\n  value={self.value}"
        res += f"\n  description={self.description}"
        res += f"\n  ref_field={self.ref_field}"
        res += f"\n  key_name={self.key_name}"
        res += "\n)"
        return res


class InstanceField(BaseField):
    """
    This is a base class. Represents an annotation of an instance that can be
    attached to any other field and can act as a label.

    Args:
        instance_id (id): Id of the instance within the dataset.
        label (AnnotationClass): An AnnotationClass associated with the instance.
        confidence (float): Confidence level of the field.
        description (Optional[str]): A description of the object or annotation information.
        ref_field (List[object]): A list of objects the current BaseField is refering to.
        key_name (Optional[str]): A string key, by which the field can be identified.
        user_attributes (Optional[Dict[str, object]]): Dictionary containing additional data regarding the field.

    Returns:
        :class: `efemarai.fields.annotation_fields.InstanceField`: An InstanceField object.
    """

    def __init__(
        self,
        instance_id: int,
        label: AnnotationClass = None,
        confidence: float = None,
        id: ObjectId = None,
        description: str = None,
        ref_field: list = None,
        key_name: str = None,
        user_attributes: Dict[str, object] = None,
    ):
        super(InstanceField, self).__init__(
            confidence=confidence,
            id=id,
            description=description,
            ref_field=ref_field,
            key_name=key_name,
            user_attributes=user_attributes,
        )

        self.instance_id = instance_id
        self.label = (
            label
            if isinstance(label, (AnnotationClass, type(None)))
            else AnnotationClass(**label)
        )


class BoundingBox(InstanceField):
    """
    Represents a bounding box annotation.

    Args:
        xyxy (Tuple[float]): Coordinates of the bounding box in absolute values in format x1,y1,x2,y2.
        label (AnnotationClass): An AnnotationClass associated with the instance
        area (Optional[float]): Area of the bounding box
        instance_id (Optional[id]): Id of the instance within the dataset
        confidence (Optional[float]): Confidence level of the field.
        description (Optional[str]): A description of the object or annotation information.
        ref_field (List[object]): A list of objects the current BaseField is refering to.
        key_name (Optional[str]): A string key, by which the field can be identified.
        user_attributes (Optional[Dict[str, object]]): Dictionary containing additional data regarding the field.

    Returns:
        :class: `efemarai.fields.annotation_fields.BoundingBox`: A BoundingBox object.
    """

    def __init__(
        self,
        xyxy: tuple,
        label: AnnotationClass,
        area: float = None,
        instance_id: int = None,
        description: str = None,
        confidence: float = None,
        id: ObjectId = None,
        ref_field: list = None,
        key_name: str = None,
        user_attributes: Dict[str, object] = None,
    ):
        super().__init__(
            instance_id=instance_id,
            label=label,
            confidence=confidence,
            id=id,
            description=description,
            ref_field=ref_field,
            key_name=key_name,
            user_attributes=user_attributes,
        )

        self.xyxy = tuple(
            i if not isinstance(i, (np.int64, np.int32)) else i.item() for i in xyxy
        )
        self.area = (
            area
            if area is not None
            else (self.xyxy[2] - self.xyxy[0]) * (self.xyxy[3] - self.xyxy[1])
        )

    def __repr__(self):
        res = f"{self.__module__}.{self.__class__.__name__}("
        res += f"\n  xyxy={self.xyxy}"
        res += f"\n  area={self.area}"
        res += f"\n  instance_id={self.instance_id}"
        res += f"\n  description={self.description}"
        res += f"\n  ref_field={self.ref_field}"
        res += f"\n  label={self.label}"
        res += f"\n  key_name={self.key_name}"
        res += "\n)"
        return res

    @property
    def width(self):
        return self.xyxy[2] - self.xyxy[0] + 1

    @property
    def height(self):
        return self.xyxy[3] - self.xyxy[1] + 1

    @property
    def size(self):
        return self.width * self.height

    @property
    def ratio(self):
        return self.width / self.height


class Polygon(InstanceField):
    """
    Represents a polygon annotation.

    Args:
        vertices List[List[Tuple[float]]]: Coordinates of the polygon. The
            values in are in format [[[point]...],[[point]...]] in image dimensions,
            where the first list contains all contours of the polygon, the internal
            ones - lists of single points of that contour.
        label (AnnotationClass): An AnnotationClass associated with the instance
        area (Optional[float]): Area of the polygon
        instance_id (Optional[id]): Id of the instance within the dataset
        confidence (Optional[float]): Confidence level of the field.
        description (Optional[str]): A description of the object or annotation information.
        ref_field (List[object]): A list of objects the current BaseField is refering to.
        key_name (Optional[str]): A string key, by which the field can be identified.
        user_attributes (Optional[Dict[str, object]]): Dictionary containing
            additional data regarding the field.

    Returns:
        :class: `efemarai.fields.annotation_fields.Polygon`: A Polygon object.

    You can construct a polygon from the output of a model that is a mask:

    .. code-block:: python

        import efemarai as ef
        import numpy as np

        mask = np.zeros((100, 100), dtype=np.uint8)
        mask[10:30, 10:50] = 255

        label = ef.AnnotationClass(id=1)
        polygon = ef.InstanceMask(label=label, data=mask).to_polygon()
        print(label, polygon)

        new_poly = ef.Polygon(label=label, vertices=polygon.vertices)
        print(new_poly)
    """

    def __init__(
        self,
        vertices: List[List[Tuple[float, float]]],
        label: AnnotationClass,
        area: float = None,
        instance_id: int = None,
        confidence: float = None,
        id: ObjectId = None,
        description: str = None,
        ref_field: list = None,
        key_name: str = None,
        user_attributes: Dict[str, object] = None,
    ):
        super().__init__(
            instance_id=instance_id,
            label=label,
            confidence=confidence,
            id=id,
            description=description,
            ref_field=ref_field,
            key_name=key_name,
            user_attributes=user_attributes,
        )

        if not (
            isinstance(vertices, (list, tuple))
            and all(
                (isinstance(v, (list, tuple)))
                and all(
                    (isinstance(p, (tuple, list)))
                    and len(p) == 2
                    and all(isinstance(f, (float, int)) for f in p)
                    for p in v
                )
                for v in vertices
            )
        ):
            raise ValueError(
                "Polygon vertices are not of type List[List[List[float/int, float/int], ...]] or Tuple[Tuple[Tuple[float/int, float/int], ...]]"
            )

        self.vertices = vertices
        self.area = area

    def __repr__(self):
        res = f"{self.__module__}.{self.__class__.__name__}("
        res += f"\n  vertices={self.vertices}"
        res += f"\n  area={self.area}"
        res += f"\n  instance_id={self.instance_id}"
        res += f"\n  description={self.description}"
        res += f"\n  label={self.label}"
        res += f"\n  confidence={self.confidence}"
        res += f"\n  ref_field={self.ref_field}"
        res += f"\n  key_name={self.key_name}"
        res += "\n)"
        return res


class Keypoint(InstanceField):
    """
    Represents a keypoint annotation.

    Args:
        x (float): X coordinate of the keypoint in absolute value.
        y (float): Y coordinate of the keypoint in absolute value.
        label (AnnotationClass): An AnnotationClass associated with the instance
        name (Optional[str]): Name of the keypoint, typically provided within a Skeleton.
        index (Optional[int]): Index of the keypoint within a Skeleton.
        occluded (Optional[bool]): Whether the keypoint is occluded.
        annotated (Optional[bool]): In case the keypoint is occluded, if it is annotated or with default coords.
        instance_id (Optional[id]): Id of the instance within the dataset
        confidence (Optional[float]): Confidence level of the field.
        description (Optional[str]): A description of the object or annotation information.
        ref_field (List[object]): A list of objects the current BaseField is refering to.
        key_name (Optional[str]): A string key, by which the field can be identified.
        user_attributes (Optional[Dict[str, object]]): Dictionary containing additional data regarding the field.

    Returns:
        :class: `efemarai.fields.annotation_fields.Keypoint`: A Keypoint object.
    """

    def __init__(
        self,
        x: float,
        y: float,
        label: AnnotationClass,
        name: str = None,
        index: int = None,
        annotated: bool = None,
        occluded: bool = None,
        instance_id: int = None,
        confidence: float = None,
        id: ObjectId = None,
        description: str = None,
        ref_field: list = None,
        key_name: str = None,
        user_attributes: Dict[str, object] = None,
    ):
        super().__init__(
            instance_id=instance_id,
            label=label,
            confidence=confidence,
            id=id,
            description=description,
            ref_field=ref_field,
            key_name=key_name,
            user_attributes=user_attributes,
        )
        self.x = x if not isinstance(i, (np.int64, np.int32)) else i.item()
        self.y = y if not isinstance(i, (np.int64, np.int32)) else i.item()
        self.name = name
        self.index = index
        self.annotated = annotated
        self.occluded = occluded

    def __repr__(self):
        res = f"{self.__module__}.{self.__class__.__name__}("
        res += f"\n  x={self.x}"
        res += f"\n  y={self.y}"
        res += f"\n  occluded={self.occluded}"
        res += f"\n  instance_id={self.instance_id}"
        res += f"\n  description={self.description}"
        res += f"\n  label={self.label}"
        res += f"\n  ref_field={self.ref_field}"
        res += f"\n  key_name={self.key_name}"
        res += f"\n  id={self.id}"
        res += "\n)"
        return res

    @staticmethod
    def chunks(lst, n):
        """Yield successive n-sized chunks from lst."""
        for i in range(0, len(lst), n):
            yield lst[i : i + n]


class Skeleton(InstanceField):
    """
    Represents a skeleton annotation. A sekeleton is a group of keypoints that
    represent an instance of an object.

    Args:
        keypoints (List[Keypoint]): List of the Keypoints in the class
        label (AnnotationClass): An AnnotationClass associated with the instance
        edges: (Optional[List[Tuple[int]]]) List of edges between keypoints. Each edge is a tuple (from_keypoint_index, to_keypoint_index) and is used for visualization.
        instance_id (Optional[id]): Id of the instance within the dataset
        confidence (Optional[float]): Confidence level of the field.
        description (Optional[str]): A description of the object or annotation information.
        ref_field (List[object]): A list of objects the current BaseField is refering to.
        key_name (Optional[str]): A string key, by which the field can be identified.
        user_attributes (Optional[Dict[str, object]]): Dictionary containing additional data regarding the field.

    Returns:
        :class: `efemarai.fields.annotation_fields.Skeleton`: A Skeleton object.


    Example of how to convert Mediapipe facemesh into the efemarai system:

    .. code-block:: python

        import efemarai as ef
        import cv2
        import mediapipe as mp
        from mediapipe.python.solutions.drawing_utils import _normalized_to_pixel_coordinates

        model = mp.solutions.face_mesh
        # >> The image is an ef.Image variable

        results = model.process(cv2.cvtColor(image.data, cv2.COLOR_BGR2RGB))
        sdk_outputs = []
        label = ef.AnnotationClass(name="face")

        # For each face in the image
        for i, face_landmarks in enumerate(results.multi_face_landmarks):
            keypoints = []
            for j, idx in enumerate(face_landmarks.landmark):
                cord = _normalized_to_pixel_coordinates(idx.x, idx.y, image.width, image.height)
                keypoints.append(
                    ef.Keypoint(
                        x=cord[0],
                        y=cord[1],
                        name=model.name[j],
                        index=j,
                        annotated=True,
                        occluded=False,
                        instance_id=i,
                        ref_field=image,
                        label=label,
                        confidence=1,
                    )
                )

            sdk_outputs.append(
                ef.Skeleton(
                    label=label,
                    keypoints=keypoints,
                    ref_field=image,
                    instance_id=i,
                    confidence=1,
                )
            )
    """

    def __init__(
        self,
        keypoints: List[Keypoint],
        label: AnnotationClass,
        edges: list = None,
        instance_id: int = None,
        confidence: float = None,
        id: ObjectId = None,
        description: str = None,
        ref_field: list = None,
        key_name: str = None,
        user_attributes: Dict[str, object] = None,
    ):
        super().__init__(
            instance_id=instance_id,
            label=label,
            confidence=confidence,
            id=id,
            description=description,
            ref_field=ref_field,
            key_name=key_name,
            user_attributes=user_attributes,
        )
        self.keypoints = keypoints
        self.edges = [
            (f, t)
            if not isinstance(f, (np.int64, np.int32))
            or not isinstance(t, (np.int64, np.int32))
            else (int(f), int(t))
            for (f, t) in edges
        ]

    def __repr__(self):
        res = f"{self.__module__}.{self.__class__.__name__}("
        res += f"\n  keypoints={self.keypoints}"
        res += "\n)"
        return res


class PolarVector:
    def __init__(self, angle: float, length: float):

        self.angle = angle
        self.length = length



from efemarai.dataset import Dataset, DatasetFormat, DatasetStage
from efemarai.fields import (
    AnnotationClass,
    BoundingBox,
    DataFrame,
    Datapoint,
    Image,
    InstanceMask,
    Keypoint,
    Mask,
    Polygon,
    Skeleton,
    Tag,
    Text,
    Value,
    Video,
    VideoFrame,
)
from efemarai.session import Session

__version__ = "0.3.4"



import json
import os
import zipfile
from enum import Enum
from time import sleep

import boto3
from appdirs import user_data_dir
from botocore.errorfactory import ClientError

from efemarai.console import console
from efemarai.fields import AnnotationClass, sdk_deserialize
from efemarai.job_state import JobState
from efemarai.problem_type import ProblemType


class DatasetFormat(str, Enum):
    """Possible dataset formats."""

    COCO = "COCO"
    ImageRegression = "ImageRegression"
    ImageNet = "ImageNet"
    TFRecord = "tfrecord"
    Custom = "Custom"

    @classmethod
    def has(cls, value):
        try:
            cls(value)
            return True
        except ValueError:
            return False

    @classmethod
    def list(cls):
        return list(map(lambda x: x.value, cls._member_map_.values()))


class DatasetStage(str, Enum):
    Train = "train"
    Validation = "validation"
    Test = "test"

    @classmethod
    def has(cls, value):
        try:
            cls(value)
            return True
        except ValueError:
            return False

    @classmethod
    def list(cls):
        return list(map(lambda x: x.value, cls._member_map_.values()))


class Dataset:
    """
    Provides dataset related functionality.
    It can be created through the :class:`efemarai.project.Project.create_dataset` method.

    Example:

    .. code-block:: python
        :emphasize-lines: 2,5

        import efemarai as ef
        dataset = ef.Session().project("Name").create_dataset(...)
        # do something else
        dataset.reload()
        if dataset.finished:
            print('Dataset has finished loading')

    Example (2):

    .. code-block:: python

        import efemarai as ef
        project = ef.Session().project("Name")
        dataset = project.dataset("Dataset Name")
        print(f"Classes: {dataset.classes}")

    """

    @staticmethod
    def create(
        project,
        name,
        stage,
        format=None,
        data_url=None,
        annotations_url=None,
        credentials=None,
        upload=None,
        num_datapoints=None,
        mask_generation=None,
        min_asset_area=None,
    ):
        """Create a dataset. A more convenient way is to use `project.create_dataset(...)`."""
        if stage is not None and data_url is not None:
            # Remote dataset
            if not DatasetFormat.has(format):
                raise AssertionError(
                    f"Dataset format '{format}' should be in {DatasetFormat.list()}."
                )

            if not DatasetStage.has(stage):
                raise AssertionError(
                    f"Dataset stage '{stage}' should be in {DatasetStage.list()}."
                )

            if name is None or data_url is None:
                return None

            if mask_generation not in (None, "Simple", "Advanced"):
                raise AssertionError("Mask generation not in (None, Simple, Advanced)")

            response = project._put(
                "api/dataset",
                json={
                    "name": name,
                    "format": format,
                    "stage": stage,
                    "data_url": data_url,
                    "annotations_url": annotations_url,
                    "access_token": credentials,
                    "upload": upload,
                    "projectId": project.id,
                },
            )
            dataset_id = response["id"]

            if upload:
                endpoint = f"api/dataset/{dataset_id}/upload"
                if annotations_url is not None:
                    project._upload(annotations_url, endpoint)
                project._upload(data_url, endpoint)
                project._post(
                    endpoint,
                    json={
                        "num_samples": num_datapoints,
                        "mask_generation": mask_generation,
                        "min_asset_area": min_asset_area,
                    },
                )

            return Dataset(
                project=project,
                name=name,
                stage=stage,
                id=dataset_id,
                format=format,
                data_url=data_url,
                annotations_url=annotations_url,
                state="NotStarted",
                classes=[],
            )
        response = project._put(
            "api/dataset",
            json={
                "name": name,
                "format": format,
                "stage": stage,
                # "data_url": data_url,
                # "annotations_url": annotations_url,
                "access_token": credentials,
                "upload": upload,
                "projectId": project.id,
                "create_only": True,
            },
        )
        dataset_id = response["id"]

        if upload:
            endpoint = f"api/dataset/{dataset_id}/upload"
            if annotations_url is not None:
                project._upload(annotations_url, endpoint)
            project._upload(data_url, endpoint)
            project._post(
                endpoint,
                json={
                    "num_samples": num_datapoints,
                    "mask_generation": mask_generation,
                    "min_asset_area": min_asset_area,
                },
            )

        # Custom dataset to be uploaded
        return Dataset(
            project=project,
            name=name,
            stage=stage,
            id=dataset_id,
            state="NotStarted",
        )

    @staticmethod
    def _parse_classes(dataset_classes):
        if dataset_classes == []:
            return []

        # Support previous versions
        id_access_key = "id" if "id" in dataset_classes[0] else "index"

        classes = []
        for c in dataset_classes:
            classes.append(
                AnnotationClass(
                    id=c[id_access_key],
                    name=c["name"],
                    category=c.get("category"),
                    color=c.get("color"),
                )
            )

        return classes

    def __init__(
        self,
        project,
        name,
        stage,
        id=None,
        format=None,
        data_url=None,
        annotations_url=None,
        state=None,
        state_message=None,
        classes=None,
        count=None,
    ):
        self.project = (
            project  #: (:class:`efemarai.project.Project`) Associated project.
        )
        self.stage = stage  #: (str) Stage of the dataset.
        self.name = name  #: (str) Name of the dataset.
        self.classes = (
            classes if classes is not None else []
        )  #: (list [str]) List of names of the classes. `None` is used for those ids that don't have a corresponding name.
        self.id = id
        self.format = format  #: (str) Format of the dataset.
        self.data_url = data_url
        self.annotations_url = annotations_url
        self.state = JobState(state)
        self.state_message = state_message
        self.count = count

    def __repr__(self):
        res = f"{self.__module__}.{self.__class__.__name__}("
        res += f"\n  id={self.id}"
        res += f"\n  name={self.name}"
        res += f"\n  format={self.format}"
        res += f"\n  stage={self.stage}"
        res += f"\n  count={self.count}" if self.count else ""
        res += f"\n  data_url={self.data_url}" if self.data_url else ""
        res += (
            f"\n  annotations_url={self.annotations_url}"
            if self.annotations_url
            else ""
        )
        res += f"\n  state={self.state}"
        res += f"\n  state_message={self.state_message}"
        res += f"\n  classes={self.classes}"
        res += "\n)"
        return res

    @staticmethod
    def deserialize_datapoint(data):
        """Returns a datapoint from the data field."""
        datapoint = sdk_deserialize(data)
        datapoint.inputs = [sdk_deserialize(json.loads(i)) for i in datapoint.inputs]
        # Dereference target ref_fields
        for i, target in enumerate(datapoint.targets):
            data = json.loads(target)
            if "ref_field" in data and len(data["ref_field"]) > 0:
                data["ref_field"] = [
                    di
                    for ref in data["ref_field"]
                    for di in datapoint.inputs
                    if str(di.id) in str(ref)
                ]
            datapoint.targets[i] = data

        datapoint.targets = [sdk_deserialize(t) for t in datapoint.targets]
        return datapoint

    def __get_data(self, skip, limit):
        response = self.project._post(
            "api/datapointsIter",
            json={
                "datasetId": self.id,
                "skip": skip,
                "limit": limit,
                "showOnlyAllowed": True,
            },
        )

        datapoints = []
        for data in response["objects"]:
            datapoint = Dataset.deserialize_datapoint(data)
            datapoints.append(datapoint)

        return datapoints, response["count"]

    def __len__(self):
        if self.count is None:
            _, datapoints_count = self.__get_data(skip=0, limit=1)
            self.count = datapoints_count

        return self.count

    def __getitem__(self, key):
        if isinstance(key, slice):
            # Note: Currently ignoring the step in the slice
            start = key.start if key.start is not None else 0
            stop = key.stop if key.stop is not None else len(self)
            datapoints, _ = self.__get_data(skip=start, limit=stop - start)
            return datapoints
        if isinstance(key, int):
            if key < 0:  # Handle negative indices
                key += len(self)
            if key < 0 or key >= len(self):
                raise IndexError(
                    f"The index ({key}) is out of range (size: {len(self)})."
                )
            datapoints, _ = self.__get_data(skip=key, limit=1)
            return datapoints[0]
        raise TypeError(f"Invalid argument type ({type(key)}).")

    @property
    def finished(self):
        """Returns true if the dataset loading has finished.

        :rtype: bool
        """
        return self.state == JobState.Finished

    @property
    def failed(self):
        """Returns true if the dataset loading has failed.

        :rtype: bool
        """
        return self.state == JobState.Failed

    @property
    def running(self):
        """Returns true if the is still being loaded - not failed or finished.

        :rtype: bool
        """
        return self.state not in (JobState.Finished, JobState.Failed)

    def add_annotation_class(self, id, name, category=None, color=None):
        annotation = AnnotationClass(id=id, name=name, category=category, color=color)
        self.classes.append(annotation)
        return annotation

    def get_annotation_class(self, *, id=None, name=None):
        if name is not None:
            annotations = list(filter(lambda x: x.name == name, self.classes))

        if id is not None:
            annotations = list(filter(lambda x: x.id == id, self.classes))

        if len(annotations) != 1:
            console.print(
                f":poop: Received not 1 annotation with name:{name}, id: {id} - '{annotations}' len({len(annotations)})."
                f"Class data: {self.classes}",
                style="red",
            )
            return None

        return annotations[0]

    def finalize(self, min_asset_area=15, mask_generation=None):
        response = self.project._put(
            f"api/dataset/{self.id}/finalize",
            json={
                "classes": [_cls._serialize() for _cls in self.classes],
                "asset": {
                    "min_asset_area": min_asset_area,
                    "mask_generation": mask_generation,
                },
            },
        )
        self.reload()

        return self

    def delete(self, delete_dependants=False):
        """
        Deletes the dataset.

        You cannot delete an object that is used in a stress test
        or a baseline (delete those first). This cannot be undone.
        """
        if self.id is None:
            return

        self.project._delete(
            f"api/dataset/{self.id}/{delete_dependants}",
        )

    def reload(self):
        """
        Reloads the dataset *in place* from the remote endpoint and return it.

        Returns:
            The updated dataset object.
        """
        if self.id is None:
            return None

        endpoint = f"api/dataset/{self.id}"
        dataset_details = self.project._get(endpoint)

        self.name = dataset_details["name"]
        self.format = dataset_details["format"]
        self.stage = dataset_details["stage"]
        self.data_url = dataset_details.get("data_url")
        self.annotations_url = dataset_details.get("annotations_url")
        self.state = JobState(dataset_details["states"][-1]["name"])
        self.state_message = dataset_details["states"][-1].get("message")
        self.classes = self._parse_classes(dataset_details["classes"])

        return self

    def download(
        self,
        num_samples=None,
        dataset_format=None,
        path=None,
        unzip=True,
        ignore_cache=False,
    ):
        """
        Download the dataset locally.

        Args:
            num_samples (int): Number of samples to download. Leave `None` for all.
            dataset_format (str): What format to download the dataset. Currently supported include `COCO`, `YOLO`, `VOC`, `ImageNet`, `CSV`.
            path (str): The path where to download the data.
            unzip (bool): Should the downloaded zip be unzipped.
            ignore_cache (bool): Force regeneration of the dataset by ignoring the cache. May lead to slower subsequent calls.
        """
        if self.id is None:
            return

        if self.running:
            console.print(":poop: Dataset has not finished loading.", style="red")
            return None

        if path is None:
            path = user_data_dir(appname="efemarai")

        path = os.path.join(path, self.id)

        if dataset_format is None:
            if self.project.problem_type == ProblemType.Classification:
                dataset_format = "imagenet"
            elif self.project.problem_type == ProblemType.ObjectDetection:
                dataset_format = "coco"
            elif self.project.problem_type == ProblemType.InstanceSegmentation:
                dataset_format = "coco"
            elif self.project.problem_type == ProblemType.Keypoints:
                dataset_format = "coco"

        if dataset_format is None:
            console.print(":poop: Unsupported problem type.", style="red")
            return None

        if not ignore_cache:
            name = os.path.join(path, f"dataset_{dataset_format}")

            if num_samples:
                name += f"_{num_samples}"

            if os.path.exists(name):
                return name

            name += ".zip"
            if os.path.exists(name + ".zip"):
                return name

        access = self.project._post(
            "api/downloadDataset",
            json={
                "id": self.id,
                "format": dataset_format,
                "num_samples": num_samples,
                "async_download": True,
            },
        )

        s3 = boto3.client(
            "s3",
            aws_access_key_id=access["AccessKeyId"],
            aws_secret_access_key=access["SecretAccessKey"],
            aws_session_token=access["SessionToken"],
            endpoint_url=access["Url"],
        )

        with console.status(f"Preparing '{self.name}' dataset download"):
            while True:
                try:
                    response = s3.head_object(
                        Bucket=access["Bucket"], Key=access["ObjectKey"]
                    )
                    size = response["ContentLength"]
                    break
                except ClientError:
                    sleep(1)

        with self.project._session._progress_bar(verbose=True) as progress:
            task = progress.add_task("Downloading dataset ", total=float(size))

            def callback(num_bytes):
                return progress.advance(task, num_bytes)

            os.makedirs(path, exist_ok=True)
            filename = os.path.join(path, os.path.basename(access["ObjectKey"]))

            s3.download_file(
                access["Bucket"], access["ObjectKey"], filename, Callback=callback
            )

        if unzip:
            with console.status("Unzipping dataset"):
                dirname = os.path.splitext(filename)[0]
                with zipfile.ZipFile(filename, "r") as f:
                    f.extractall(dirname)

                os.remove(filename)

                filename = dirname

        console.print(
            (f":heavy_check_mark: Downloaded '{self.name}' dataset to\n  {filename}"),
            style="green",
        )

        return filename

    def enhance(self, domain, samples_per_datapoint, new_name=None):
        response = self.project._post(
            "api/datasetEnhance",
            json={
                "datasetId": self.id,
                "domainId": domain.id,
                "samplesPerDatapoint": samples_per_datapoint,
                "newName": new_name,
            },
        )
        new_dataset_id = response["value"]
        return self.project.dataset(new_dataset_id)



import itertools
import os
from copy import copy

import numpy as np
from pycocotools import mask as coco_mask
from pycocotools.coco import COCO
from tqdm import tqdm

from efemarai.dataset import DatasetFormat
from efemarai.fields import (
    BoundingBox,
    Datapoint,
    Image,
    Keypoint,
    Polygon,
    Skeleton,
    create_polygons_from_mask,
)
from efemarai.formats.config import CLASS_COLORS


def create_mask(size, annotation):
    mask = np.zeros(size[:2], np.uint8)
    if type(annotation["segmentation"]["counts"]) is list:
        # uncompressed RLE
        rle = coco_mask.frPyObjects(
            annotation["segmentation"], mask.shape[0], mask.shape[1]
        )
    else:
        # plain RLE
        rle = annotation["segmentation"]

    mask = coco_mask.decode(rle) * 255
    return mask


def create_coco_dataset(
    project,
    name,
    stage,
    data_url,
    annotations_url,
    num_datapoints,
    mask_generation,
    min_asset_area,
):
    """Upload a COCO style dataset in the system."""
    dataset = project.create_dataset(
        name=name, stage=stage, format=DatasetFormat.Custom
    )

    coco = COCO(annotations_url)

    for i, category in enumerate(coco.loadCats(coco.getCatIds())):
        dataset.add_annotation_class(
            id=category["id"],
            name=category["name"],
            category=category.get("supercategory", "supercategory"),
            color=CLASS_COLORS[i % len(CLASS_COLORS)],
        )

    for cocoImg in tqdm(coco.loadImgs(coco.getImgIds())[:num_datapoints]):
        attr_data = copy(cocoImg)
        attr_data.pop("width", None)
        attr_data.pop("height", None)
        image = Image(
            file_path=os.path.join(data_url, cocoImg["file_name"]),
            width=cocoImg["width"],
            height=cocoImg["height"],
            user_attributes={"original_data": attr_data},
        )

        datapoint = Datapoint(
            dataset=dataset,
            inputs={"image": image},
        )

        annotations = coco.loadAnns(coco.getAnnIds(imgIds=cocoImg["id"]))
        for instance_id, cocoAnn in enumerate(annotations):
            # "category_id", "bbox", "area", "segmentation", "keypoints"

            # implicitly ensures only registered classes are added
            label = dataset.get_annotation_class(id=cocoAnn["category_id"])

            # or in case class name is provided
            # label = dataset.get_annotation_class(name=annotation["category_name"]),

            if "bbox" in cocoAnn:
                datapoint.add_target(
                    BoundingBox(
                        xyxy=[
                            cocoAnn["bbox"][0],
                            cocoAnn["bbox"][1],
                            cocoAnn["bbox"][0] + cocoAnn["bbox"][2],
                            cocoAnn["bbox"][1] + cocoAnn["bbox"][3],
                        ],
                        label=label,
                        instance_id=instance_id,
                        ref_field=image,
                    )
                )

            if "segmentation" in cocoAnn:
                if isinstance(cocoAnn["segmentation"], dict):
                    # RLE encoding -> create polygon
                    mask_img = np.zeros((cocoImg["width"], cocoImg["height"]))
                    mask_img = create_mask(mask_img.shape, cocoAnn)
                    polygons, polygons_area = create_polygons_from_mask(mask_img)
                    polygons = [list(itertools.chain(*polygon)) for polygon in polygons]
                    cocoAnn["segmentation"] = polygons
                    cocoAnn["area"] = polygons_area

                vertices = []
                for polygon in cocoAnn["segmentation"]:
                    vertices.append(list(zip(polygon[:-1][::2], polygon[1:][::2])))

                datapoint.add_target(
                    Polygon(
                        label=label,
                        vertices=vertices,
                        area=cocoAnn["area"],
                        instance_id=instance_id,
                        ref_field=image,
                    )
                )

            # The person category has fully annotated skeleton which looks like
            # "id": 1,
            # "name": "person",
            # "keypoints": [
            #     "nose","left_eye","right_eye","left_ear","right_ear",
            #     "left_shoulder","right_shoulder","left_elbow","right_elbow",
            #     "left_wrist","right_wrist","left_hip","right_hip",
            #     "left_knee","right_knee","left_ankle","right_ankle"
            # ],
            # "skeleton": [
            #     [16,14],[14,12],[17,15],[15,13],[12,13],[6,12],[7,13],[6,7],
            #     [6,8],[7,9],[8,10],[9,11],[2,3],[1,2],[1,3],[2,4],[3,5],[4,6],[5,7]
            # ]
            # Add single skeleton if info available
            cocoCat = coco.loadCats(label.id)[0]
            if "skeleton" in cocoCat and "keypoints" in cocoCat:
                keypoints = [
                    Keypoint(
                        x=x,
                        y=y,
                        name=cocoCat["keypoints"][i],
                        index=i,
                        annotated=v > 0,
                        occluded=v == 1,
                        instance_id=instance_id,
                        ref_field=image,
                        label=label,
                    )
                    for i, (x, y, v) in enumerate(
                        Keypoint.chunks(cocoAnn["keypoints"], 3)
                    )
                ]

                datapoint.add_target(
                    Skeleton(
                        label=label,
                        keypoints=keypoints,
                        edges=cocoCat["skeleton"],
                        ref_field=image,
                        instance_id=instance_id,
                    )
                )
            elif "keypoints" in cocoAnn:
                # else just add multiple keypoints, without grouping them
                for i, (x, y, v) in enumerate(Keypoint.chunks(cocoAnn["keypoints"], 3)):
                    datapoint.add_target(
                        Keypoint(
                            label=label,
                            x=x,
                            y=y,
                            occluded=v == 1,
                            annotated=v > 0,
                            index=i,
                            instance_id=instance_id,
                            ref_field=image,
                        )
                    )

        datapoint.upload()

    dataset.finalize(mask_generation=mask_generation, min_asset_area=min_asset_area)
    return dataset



import hashlib
import os
import tempfile
from contextlib import contextmanager
from glob import glob
from pathlib import Path
from zipfile import ZipFile

import git
import pathspec

from efemarai.console import console
from efemarai.definition_checker import DefinitionChecker


class ModelRepository:
    def __init__(self, url=None, branch=None, hash=None, access_token=None):
        self.url = url if url is not None else "."
        self.branch = branch
        self.hash = hash
        self.access_token = access_token
        self.files = self.get_all_files() if not self.is_remote else []

    def __repr__(self):
        res = f"{self.__module__}.{self.__class__.__name__}("
        res += f"\n      url={self.url}"
        res += f"\n      branch={self.branch}"
        res += "\n    )"
        return res

    @property
    def is_remote(self):
        return DefinitionChecker.is_path_remote(self.url)

    def get_all_files(self):
        ignore_files = [
            ignore_file
            for ignore_file in [".gitignore", ".efignore", ".efemaraiignore"]
            if Path.joinpath(Path(self.url), ignore_file).exists()
        ]

        # Do not upload .git folder
        ignore_lines = [".git"]

        # Filter out files based on *ignore specification
        for ignore_file in ignore_files:
            ignore_lines.extend(Path(ignore_file).read_text().splitlines())

        ignore_spec = pathspec.PathSpec.from_lines("gitwildmatch", ignore_lines)

        files = [
            file
            for file in list(Path(self.url).glob("**/*"))
            if not ignore_spec.match_file(str(file))
        ]

        return files

    @staticmethod
    def calculate_file_hash(filepath):
        return ModelHasher._calculate_file_hash(filepath=filepath)

    def get_last_commit(self, access_token):
        remote_heads = git.cmd.Git().ls_remote(
            self.url.replace(
                "//github",
                f"//{access_token}:x-oauth-basic@github",
            ),
            heads=True,
        )
        return remote_heads.split(f"\trefs/heads/{self.branch}")[0]

    @contextmanager
    def archive(self, name):
        with tempfile.TemporaryDirectory() as dirpath:
            zip_name = os.path.join(dirpath, name)

            with ZipFile(zip_name, "w") as zip:
                for file in self.files:
                    zip.write(file)

            filesize_gb = os.stat(zip_name).st_size / (1024**3)
            if filesize_gb > 1:
                console.print(
                    f":face_with_monocle: "
                    f"Model code is suspiciously large ({filesize_gb:.2f} GB) - "
                    f"consider using '.gitignore' or '.efignore' file",
                    style="orange1",
                )

            yield zip_name


class ModelFile:
    def __init__(self, name, url, upload=True, credentials=None, hash_code=None):
        self.name = name
        self.url = url
        self.upload = upload
        self.credentials = credentials
        self.hash_code = hash_code

    def __repr__(self):
        res = f"{self.__module__}.{self.__class__.__name__}("
        res += f"\n      name={self.name}"
        res += f"\n      url={self.url}"
        res += f"\n      upload={self.upload}"
        res += f"\n      hash_code={self.hash_code}"
        res += "\n    )"
        return res


class ModelHasher:
    @staticmethod
    def _calculate_file_hash(filepath, hashing_algorithm="sha1"):
        """This function returns the SHA-1 hash
        of the file passed into it."""
        # make a hash object
        if hashing_algorithm == "md5":
            h = hashlib.md5()
        elif hashing_algorithm == "sha1":
            h = hashlib.sha1()

        if str(filepath).startswith("http"):
            return hashlib.sha1(str(filepath).encode("utf-8")).hexdigest()

        if not os.path.exists(filepath):
            console.print(
                f":poop: File/folder under '{filepath}' does not exist.", style="red"
            )
            raise ValueError()

        if os.path.isdir(filepath):
            filenames = [
                filename
                for filename in glob(os.path.join(filepath, "**/*"), recursive=True)
                if os.path.isfile(filename)
            ]
            file_hashes = [ModelHasher._calculate_file_hash(file) for file in filenames]
            h_hash = b"".join(f"{f}".encode("utf-8") for f in file_hashes)
            h.update(h_hash)
            return h.hexdigest()

        # open file for reading in binary mode
        with open(filepath, "rb") as file:
            # loop till the end of the file
            chunk = 0
            while chunk != b"":
                # read only 1024 bytes at a time
                chunk = file.read(1024)
                h.update(chunk)

        # return the hex representation of digest
        return h.hexdigest()


class Model:
    """
    Provides model related functionality.

    It should be created through the :class:`efemarai.project.Project.create_model` method.

    Example:

    .. code-block:: python
        :emphasize-lines: 2

        import efemarai as ef
        ef.Session().project("Name").create_model(...)
    """

    @staticmethod
    def create(project, name, description, repository, files):
        """
        Creates a model.

        You should use :func:`project.create_model` instead.
        """
        if name is None:
            raise ValueError("Missing model name")

        if description is None:
            description = ""

        if repository is None:
            repository = {}

        if files is None:
            files = []

        if not isinstance(repository, ModelRepository):
            repository = ModelRepository(**repository)

        for f in files:
            f["hash_code"] = ModelHasher._calculate_file_hash(filepath=f["url"])

        model_files_hashes = "".join([f["hash_code"] for f in files])

        # Remote repo case
        if repository.is_remote:
            repo_hash = (
                repository.hash
                if repository.hash is not None
                else repository.get_last_commit(repository.access_token)
            )
        else:
            # Local repo case
            if len(repository.files) == 0:
                console.print(
                    f"No files in '{repository.url}'. "
                    "Confirm path and ignore rules in '.gitignore', '.efignore'."
                )
                raise ValueError()

            repo_hash = "".join(
                [
                    ModelHasher._calculate_file_hash(filepath=i)
                    for i in repository.files
                    if i.is_file()
                ]
            )

        version = hashlib.sha1(
            (repo_hash + model_files_hashes).encode("utf-8")
        ).hexdigest()  # Model hash

        files = [ModelFile(**f) if not isinstance(f, ModelFile) else f for f in files]

        response = project._put(
            "api/model",
            json={
                "name": name,
                "description": description,
                "version": version,
                "repository": {
                    "url": repository.url,
                    "branch": repository.branch,
                    "hash": repository.hash,
                    "access_token": repository.access_token,
                },
                "files": [
                    {
                        "name": f.name,
                        "url": f.url,
                        "upload": f.upload,
                        "credentials": f.credentials,
                        "hash": f.hash_code,
                    }
                    for f in files
                ],
            },
        )
        model_id = response["id"]

        if not repository.is_remote:
            base_name = "model_code.zip"
            endpoint = f"api/modelCode/{model_id}/upload"
            with repository.archive(name=base_name) as archive_name:
                project._upload(archive_name, endpoint)

            project._post(endpoint, json={"archive": base_name})

        for f in files:
            if f.upload:
                project._upload(f.url, f"api/model/{model_id}/upload")

        return Model(project, model_id, name, description, version, repository, files)

    @staticmethod
    def push(project, name, description, repository, files, existing_model):
        """
        Pushes a new model version.

        You should use :func:`project.push_model` instead.
        """
        if name is None:
            raise ValueError("Missing model name")

        if description is None:
            description = ""

        if repository is None:
            repository = {}

        if files is None:
            files = []

        if not isinstance(repository, ModelRepository):
            repository = ModelRepository(**repository)

        for f in files:
            f["hash_code"] = ModelHasher._calculate_file_hash(filepath=f["url"])

        model_files_hashes = "".join([f["hash_code"] for f in files])

        if repository.is_remote:
            # Remote repo
            repo_hash = (
                repository.hash
                if repository.hash is not None
                else repository.get_last_commit(repository.access_token)
            )
        else:
            # Local repo
            if len(repository.files) == 0:
                console.print(
                    f"No files in '{repository.url}'. "
                    "Confirm path and ignore rules in '.gitignore', '.efignore'."
                )
                raise ValueError()

            repo_hash = "".join(
                [
                    ModelHasher._calculate_file_hash(filepath=f)
                    for f in repository.files
                    if f.is_file()
                ]
            )

        version = hashlib.sha1(
            (repo_hash + model_files_hashes).encode("utf-8")
        ).hexdigest()  # Model hash

        # Do not create a new model version if there are no changes
        if version == existing_model.version:
            console.print(
                ":face_with_monocle: "
                "There are were no changes found in the files. Skipping...",
                style="orange1",
            )
            return existing_model

        files = [ModelFile(**f) if not isinstance(f, ModelFile) else f for f in files]
        files = Model._point_to_existing_model_files(
            new_model_files=files, existing_model=existing_model
        )
        response = project._put(
            "api/model",
            json={
                "name": name,
                "description": description,
                "version": version,
                "repository": {
                    "url": repository.url,
                    "branch": repository.branch,
                    "hash": repository.hash,
                    "access_token": repository.access_token,
                },
                "files": [
                    {
                        "name": f.name,
                        "url": f.url,
                        "upload": f.upload,
                        "credentials": f.credentials,
                        "hash": f.hash_code,
                    }
                    for f in files
                ],
            },
        )
        model_id = response["id"]

        if not repository.is_remote:
            base_name = "model_code.zip"
            endpoint = f"api/modelCode/{model_id}/upload"
            with repository.archive(name=base_name) as archive_name:
                project._upload(archive_name, endpoint)

            project._post(endpoint, json={"archive": base_name})

        for f in files:
            if f.upload:
                project._upload(f.url, f"api/model/{model_id}/upload")

        return Model(project, model_id, name, description, version, repository, files)

    def __init__(self, project, id, name, description, version, repository, files):
        self.project = project
        self.id = id
        self.name = name
        self.description = description
        self.version = version
        self.repository = repository
        self.files = files

    def __repr__(self):
        res = f"{self.__module__}.{self.__class__.__name__}("
        res += f"\n  id={self.id}"
        res += f"\n  name={self.name}"
        res += f"\n  description={self.description}"
        res += f"\n  version={self.version}"
        res += f"\n  repository={self.repository}"
        res += f"\n  files={self.files}"
        res += "\n)"
        return res

    def delete(self, delete_dependants=False):
        """
        Deletes the model.

        You cannot delete an object that is used in a stress test or a baseline
        (delete those first). Deletion cannot be undone.
        """
        self.project._delete(
            f"api/model/{self.id}/{delete_dependants}",
        )

    @staticmethod
    def _point_to_existing_model_files(new_model_files, existing_model):
        """
        Checks if the new model files are different from the old ones.
        Points the url to the old ones if they are the same.
        """
        for f in new_model_files:
            if f.hash_code in [f.hash_code for f in existing_model.files]:
                f.url = [
                    file.url
                    for file in existing_model.files
                    if file.hash_code == f.hash_code
                ][0]
                f.upload = False

        return new_model_files

    def versions(self):
        """Returns all model versions."""
        return self.project._get(f"api/modelVersions/{self.id}")



import glob
import os
from itertools import islice

import cv2

from efemarai.dataset import DatasetFormat
from efemarai.fields import Datapoint, Image, Tag
from efemarai.formats.config import CLASS_COLORS, IMAGE_EXTENSIONS


def get_image_width_height(image_path):
    img = cv2.imread(image_path)
    return img.shape[1], img.shape[0]


def create_imagenet_dataset(
    project,
    name,
    stage,
    data_url,
    num_datapoints,
):
    """Upload an ImageNet style dataset in the system."""
    dataset = project.create_dataset(
        name=name, stage=stage, format=DatasetFormat.Custom
    )

    # Creating classes based on folders
    classes = sorted(next(os.walk(data_url.replace("file://", "", 1)))[1])

    for i, class_name in enumerate(classes):
        dataset.add_annotation_class(
            id=i,
            name=class_name,
            color=CLASS_COLORS[i % len(CLASS_COLORS)],
        )

    # Loading images into datapoints for each folder
    for class_id, class_name in enumerate(classes):
        for ext in IMAGE_EXTENSIONS:
            for image_file in islice(
                glob.iglob(
                    f"{data_url.replace('file://', '', 1)}/{class_name}/*.{ext}"
                ),
                num_datapoints,
            ):
                label = dataset.get_annotation_class(id=class_id)

                width, height = get_image_width_height(image_file)

                image = Image(
                    file_path=image_file,
                    width=width,
                    height=height,
                    user_attributes={
                        "extension": ext,
                        "original_data": os.path.join(class_name, image_file),
                    },
                )
                tag = Tag(label=label, ref_field=image)

                datapoint = Datapoint(
                    dataset=dataset, inputs={"image": image}, targets={"class": tag}
                )
                datapoint.upload()

    dataset.finalize()

    return dataset



import inspect
import sys

import numpy as np

from efemarai.fields.annotation_fields import (
    AnnotationClass,
    BoundingBox,
    Keypoint,
    Polygon,
    Skeleton,
    Tag,
    Value,
)
from efemarai.fields.base_fields import (
    BaseField,
    _decode_ndarray,
    _encode_ndarray,
    sdk_serialize,
)
from efemarai.fields.data_fields import (
    DataFrame,
    Image,
    InstanceMask,
    Mask,
    Text,
    Video,
    VideoFrame,
    create_polygons_from_mask,
)
from efemarai.fields.datapoint import Datapoint

SDK_CLASS_STRUCTURE = dict(inspect.getmembers(sys.modules[__name__], inspect.isclass))


def sdk_deserialize(data):
    if "_cls" not in data:
        print(f":poop: Cannot locate class from the data {data.keys()}.")
        return None

    c = SDK_CLASS_STRUCTURE[data["_cls"]]
    sdk_type = data.pop("_cls")

    # "data" stores the loaded image in an np.ndarray
    if (
        "data" in data
        and isinstance(data["data"], dict)
        and all(k in data["data"].keys() for k in ("data", "shape", "dtype"))
    ):
        data["data"] = _decode_ndarray(data["data"])

    # Find id
    if "_id" in data:
        if isinstance(data["_id"], dict):
            # Get only the oid value, not the whole dict
            data["id"] = data["_id"]["$oid"]
        else:
            data["id"] = data["_id"]
        data.pop("_id")
    elif "id" in data:
        if isinstance(data["id"], dict):
            data["id"] = data["id"]["$oid"]
    else:
        print(f":poop: Cannot resolve id from data. ({data.keys()})")

    if sdk_type == "Video":
        for frame in data["frames"]:
            frame.pop("_raw", None)
            frame.pop("_cls", None)
            frame["id"] = frame["id"]["$oid"]  # TODO: Find a smarter way to do this
        data["frames"] = [VideoFrame(**f) for f in data["frames"]]

    try:
        obj = c(**data)
    except Exception:
        print(
            f"Cannot create an instance of class '{c}' with keys {data.keys()} and data {data}"
        )
    return obj



import json
import os
import tempfile
from glob import glob
from typing import Union

import ffmpeg
from bson.objectid import ObjectId

from efemarai.console import console
from efemarai.fields.annotation_fields import InstanceField, Polygon
from efemarai.fields.base_fields import BaseField
from efemarai.fields.data_fields import Video, VideoFrame


class Datapoint:
    """
    Used to represent the information about a single piece of data, which is
    to be passed to the model. The datapoint holds the data in `inputs` (like
    images, text, etc) and the annotations (bounding boxes, polygons, etc).

    Args:
        dataset (efemarai.dataset.Dataset): Instance of a Dataset object
        inputs (list[BaseFields]): List of Basefields to represent the inputs to the datapoint.
        targets (list[BaseFields]): List of Basefields to represent the targets to the inputs datapoint.
        id (Optional[ObjectId]): A valid ObjectId for the datapoint
    Returns:
        :class: `efemarai.fields.datapoint.Datapoint`
    """

    def __init__(self, dataset, inputs=None, targets=None, id=None) -> None:
        self.inputs = self.sanitize_format(inputs)
        self.targets = self.sanitize_format(targets)
        self.dataset = dataset
        self.id = id if id is not None else ObjectId()

    def __repr__(self):
        res = f"{self.__module__}.{self.__class__.__name__}("
        res += f"\n  id={self.id}"
        res += f"\n  dataset={self.dataset}"
        res += f"\n  inputs={self.inputs}"
        res += f"\n  targets={self.targets}"
        res += "\n)"
        return res

    @staticmethod
    def sanitize_format(data):
        if data is None:
            return []

        if isinstance(data, list):
            return data

        if isinstance(data, dict):
            for k, v in data.items():
                v.key_name = k
            return list(data.values())
        raise AttributeError("Datapoint inputs and targets should be list or dict.")

    def add_target(self, _target: Union[BaseField, InstanceField]):
        """
        Used to add target to the datapoint.

        Args:
            _target: An instance of efemarai.fields.base_fields.BaseField.

        Returns:
            targets: list[BaseField]

        """
        if isinstance(_target, Polygon) and (
            not isinstance(_target.vertices, (list, tuple))
            or not len(_target.vertices) > 0
            or not isinstance(_target.vertices[0][0][0], (float, int))
        ):
            console.print(
                ":poop: We do not support empty polygons or RLE style polygons yet. Please pass a list of lists of floats.",
                style="red",
            )
            return None

        self.targets.append(_target)
        return self.targets

    def add_inputs(self, _input: Union[BaseField, InstanceField]):
        """
        Used to add inputs to the datapoint.

        Args:
            _input: An instance of efemarai.fields.base_fields.BaseField.

        Returns:
            inputs: list[BaseField]

        """
        self.inputs.append(_input)
        return self.inputs

    def _post(self, endpoint, json=None, params=None):
        return self.dataset.project._session._post(endpoint, json, params)

    def _put(self, endpoint, json=None, params=None):
        return self.dataset.project._put(endpoint, json, params)

    def _handle_video(self, ef_video):
        with tempfile.TemporaryDirectory() as TMP_VIDEO_FRAME_DIR:
            video_info = ffmpeg.probe(ef_video.file_path)["streams"]
            width = [info["width"] for info in video_info if info.get("width")][0]
            height = [info["height"] for info in video_info if info.get("height")][0]

            basename = os.path.basename(ef_video.file_path)
            out_path = os.path.join(TMP_VIDEO_FRAME_DIR, basename)
            os.makedirs(out_path, exist_ok=True)
            ffmpeg.input(ef_video.file_path).output(
                f"{out_path}/{basename}_frame_%08d.png", **{"qscale:v": 2}
            ).run(quiet=True)

            frames = sorted(glob(f"{out_path}/{basename}_frame_*.png"))
            # ffmpeg r_frame_rate is a string in decimal format -> "12/1"
            r_frame_rate = [
                info["r_frame_rate"] for info in video_info if info.get("width")
            ][0].split("/")
            fps = int(r_frame_rate[0]) / int(r_frame_rate[1])
            ef_video.width = width
            ef_video.height = height
            ef_video.fps = fps
            ef_video.frame_count = len(frames)
            for frame in frames:
                ef_video.frames.append(
                    VideoFrame(
                        index=frame.split("_")[-1].strip(".png"),
                        file_path=frame,
                        width=width,
                        height=height,
                    )
                )
                self.dataset.project._upload(
                    frame,
                    f"api/datapoint/{self.dataset.id}/upload",
                    verbose=False,
                )
                os.remove(frame)

    def upload(self, copy_files=True):
        """
        Used to upload a datapoint to the system.

        Args:
            copy_files (bool): Whether to copy the files or to upload them

        Returns:
            datapoint_id: ObjectId of the uploaded datapoint
        """
        files_uploaded = []
        # Upload files associated with inputs and targets
        for field in self.inputs + self.targets:
            if hasattr(field, "file_path") and isinstance(field.file_path, str):
                if not os.path.isfile(field.file_path):
                    console.print(
                        f":poop: Expecting a file at '{field.file_path}', but cannot locate it.",
                        style="red",
                    )
                    continue

                # Uploaded this file, so skipping
                if field.file_path in files_uploaded:
                    continue

                self.dataset.project._upload(
                    field.file_path,
                    f"api/datapoint/{self.dataset.id}/upload",
                    verbose=False,
                )

                files_uploaded.append(field.file_path)

                if isinstance(field, Video):
                    self._handle_video(field)

        serialized_targets = [
            json.loads(target._serialize()) for target in self.targets
        ]
        serialized_inputs = [json.loads(_input._serialize()) for _input in self.inputs]

        response = self._put(
            "api/datapoint/undefined",
            json={
                "access_token": self.dataset.project._session.token,
                "datasetId": self.dataset.id,
                "targets": serialized_targets,
                "inputs": serialized_inputs,
            },
        )

        datapoint_id = response["id"]
        if not copy_files:
            return datapoint_id

        return datapoint_id

    def get_input(self, name):
        """
        Returns an input with that particular key name.

        Args:
            name (str): Key name of a specific field to be returned, or None if not found.

        Returns:
            An instance of BaseField
        """
        for _input in self.inputs:
            if _input.key_name == name:
                return _input

        return None

    def get_inputs_with_type(self, ef_type):
        """
        A way to filter only inputs of a particular ef_type.

        Args:
            ef_type (BaseField): An instance of `efemarai.fields.base_fields.BaseField`.

        Returns:
            list[BaseField] - a list of inputs of the ef_type

        Example:

        .. code-block:: python

            import efemarai as ef

            images = datapoints[0].get_inputs_with_type(ef.Image)
            # images is a list of all of the ef.Image elements in datapoints[0].
            # each of those images can be accessed by datapoints[0].get_input(images[0].key_name)
        """
        res = []
        for _input in self.inputs:
            if ef_type is type(Video):
                res.append(_input)
                continue
            if isinstance(_input, ef_type):
                res.append(_input)
            if _input._cls == "Video":
                # TODO: Use frame.index when indexes start from 0. Currently they start at 1.
                for idx, frame in enumerate(_input.frames):
                    frame.set_data(_input.data[idx])
                    res.append(frame)
        return res

    def get_targets_with_type(self, ef_type):
        """
        A way to filter only targets of a particular ef_type.

        Args:
            ef_type (BaseField): An instance of `efemarai.fields.base_fields.BaseField`.

        Returns:
            list[BaseField] - a list of inputs of the ef_type
        """
        res = []
        for target in self.targets:
            if isinstance(target, ef_type):
                res.append(target)
        return res

    def get_input_refs(self, ids):
        """
        Returns a list of field in input that refer to id.

        Args:
            ids (List[ObjectId]): Ids when referenced to return object, or [] if not found.

        Returns:
            An instance of BaseField
        """
        refs = []
        for _input in self.inputs:
            if _input.id in ids:
                refs.append(_input)

        return refs



import base64
import json
from typing import Dict, List

import numpy as np
from bson.objectid import ObjectId


def _decode_ndarray(xs):
    data = base64.b64decode(xs["data"])
    shape = xs["shape"]
    dtype = np.dtype(xs["dtype"])
    return np.frombuffer(data, dtype=dtype).reshape(*shape)


def _encode_ndarray(xs):
    data = xs.data if xs.flags["C_CONTIGUOUS"] else xs.tobytes()
    return {
        "data": base64.b64encode(data).decode("ascii"),
        "shape": xs.shape,
        "dtype": str(xs.dtype),
    }


def sdk_serialize(obj):
    return json.dumps(obj, default=_serialize_function)


def _serialize_function(x):
    if isinstance(x, np.ndarray):
        res = _encode_ndarray(x)
        return res
    if isinstance(x, ObjectId):
        return str(x)
    return x.__dict__


class BaseField:
    """
    At the core of each data field, the BaseField creates a common representation of an
    element in Efemarai. It is used to store the id, description, reference field, name,
    confidence, and any user added attributes.

    Args:
        id (ObjectId): A BaseField unique identifier.
        description (str): A description of the object or annotation information.
        ref_field (List[object]): A list of objects the current BaseField is refering.
        key_name (str): A string key, by which the field can be identified.
        confidence (float): Confidence level of the field.
        user_attributes (Dict[str, object]): Optional dictionary containing additional
            data regarding the field.

    Returns:
        :class:`efemarai.fields.base_field.BaseField`: A BaseField object.
    """

    def __init__(
        self,
        id: ObjectId = None,
        description: str = None,
        ref_field: List[object] = None,
        key_name: str = None,
        confidence: float = None,
        user_attributes: Dict[str, object] = None,
    ):

        self.id = id if id is not None else ObjectId()
        self.description = description
        self.key_name = key_name
        self.confidence = confidence
        self._cls = self.__class__.__name__
        self.ref_field = ref_field
        self.user_attributes = user_attributes

        # Propagate ref_field to frames in video
        if ref_field is not None:
            if not isinstance(ref_field, list):
                if hasattr(ref_field, "_cls") and ref_field._cls == "Video":
                    self.frames = (
                        [ref_field._index] if ref_field._index is not None else []
                    )
                self.ref_field = [
                    ref_field.id if hasattr(ref_field, "id") else ref_field
                ]
            else:
                for ref in ref_field:
                    if hasattr(ref, "_cls") and ref._cls == "Video":
                        if not hasattr(self, "frames") or not self.frames:
                            self.frames = []
                        self.frames.append([ref._index for ref in ref_field])
                self.ref_field = [
                    ref.id if hasattr(ref, "id") else ref for ref in ref_field
                ]

    def _serialize(self):
        return sdk_serialize(self)

    def __repr__(self):
        res = f"{self.__module__}.{self.__class__.__name__}("
        res += f"\n  id={self.id}"
        res += f"\n  description={self.description}"
        res += f"\n  ref_field={self.ref_field}"
        res += f"\n  key_name={self.key_name}"
        res += f"\n  confidence={self.confidence}"
        res += f"\n  user_attributes={self.user_attributes}"
        res += "\n)"
        return res



IMAGE_EXTENSIONS = [
    "png",
    "jpg",
    "jpeg",
    "gif",
    "tiff",
    "bmp",
    "webp",
]

# These colors follow the tab20 colormap, however it's been rearranged
# such that brighter, more modern looking colors are at the beginning
CLASS_COLORS = [
    # Dark
    "#1f77b4",  # blue
    "#ff7f0e",  # orange
    "#bcbd22",  # lime
    "#17becf",  # cyan
    "#9467bd",  # purple
    "#8c564b",  # brown
    "#e377c2",  # pink
    "#7f7f7f",  # grey
    "#2ca02c",  # green
    "#d62728",  # red
    # Light
    "#aec7e8",  # blue
    "#ffbb78",  # orange
    "#dbdb8d",  # lime
    "#9edae5",  # cyan
    "#c5b0d5",  # purple
    "#c49c94",  # brown
    "#f7b6d2",  # pink
    "#c7c7c7",  # grey
    "#98df8a",  # green
    "#ff9896",  # red
]



import itertools
import math
from typing import Dict

import cv2
import numpy as np
from bson.objectid import ObjectId
from PIL import Image as pil_image

from efemarai.fields.annotation_fields import AnnotationClass, InstanceField, Polygon
from efemarai.fields.base_fields import BaseField, sdk_serialize


def create_polygons_from_mask(mask_img, threshold_value=127):
    # Get contours as polygons and the area of the polygons
    ret, thresh = cv2.threshold(mask_img, threshold_value, 255, 0)
    (
        contours,
        hierarchy,
    ) = cv2.findContours(  # Format: [[[[x1, y1]], [[x2, y2]]...], ...]
        thresh, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE
    )
    if not contours:
        return ([], 0)

    polygons_area = 0
    polygons = []  # Format: [[x1, y1, x2, y2...], []]

    for contour in contours:
        # Skip single points and lines
        if len(contour) < 3:
            continue

        polygons_area += cv2.contourArea(contour)

        # Get rid of unnecessary levels of nesting
        points = list(itertools.chain(*contour))  # Format: [[x1, y1], [x2, y2]...]

        polygons.append([[float(coord) for coord in point] for point in points])

    return (polygons, polygons_area)


class Image(BaseField):
    """
    Represents an image that can be attached to any other field and can
    act as an input or a target. Either the `file_path` or the `data` field
    needs to be passed to be a valid element.

    Args:
        file_path (str): The location of the image to be loaded.
        data (np.ndarray): The numpy array with the loaded image of type `uint8`.
        width (Optional[int]): Width of the image (autofield if data is passed).
        height (Optional[int]): Height of the image (autofield if data is passed).
        description (Optional[str]): A description of the object or annotation information.
        ref_field (Optional[List[object]]): A list of objects the current BaseField is refering.
        key_name (Optional[str]): A string key, by which the field can be identified.
        confidence (Optional[float]): Confidence level of the field.
        user_attributes (Optional[Dict[str, object]]): Optional dictionary containing additional
            data regarding the field.

    Returns:
        :class: `efemarai.fields.data_fields.Image`: An Image object.

    Example:

    .. code-block:: python

        import efemarai as ef
        import numpy as np

        image = np.zeros((100, 100), dtype=np.uint8)
        image[10:30, 10:50] = 255

        ef_image = ef.Image(data=image)
        print(ef_image)

        ef_image2 = ef.Image(file_path="./image.png")
        print(ef_image2)
    """

    def __init__(
        self,
        file_path: str = None,
        data: np.ndarray = None,
        width: int = None,
        height: int = None,
        confidence: float = None,
        id: ObjectId = None,
        description: str = None,
        ref_field: list = None,
        key_name: str = None,
        user_attributes: Dict[str, object] = None,
    ):
        BaseField.__init__(
            self,
            confidence=confidence,
            id=id,
            description=description,
            ref_field=ref_field,
            key_name=key_name,
            user_attributes=user_attributes,
        )

        if file_path is None and data is None:
            raise AssertionError(
                f"You need to specify either a 'file_path'({file_path}) or 'data'({data})."
            )

        self.file_path = file_path
        self.width = width
        self.height = height
        self._raw = data

        # Skip check for 0-1 images (TODO: give warning)
        if isinstance(self._raw, list):
            self._raw = np.asarray(self._raw, dtype=np.uint8)
        if self._raw is not None and self._raw.dtype != np.uint8:
            raise AttributeError(
                f"Expected format {np.uint8}. Received {self._raw.dtype}."
            )

        if height is None:
            if self._raw is not None:
                self.height = self._raw.shape[0]
            else:
                img = cv2.imread(self.file_path, cv2.IMREAD_UNCHANGED)
                self.height = img.shape[0]
                self.width = img.shape[1]

        if width is None and self._raw is not None:
            self.width = self._raw.shape[1]

    @property
    def data(self):
        if self._raw is None:
            self._raw = np.asarray(pil_image.open(self.file_path))
            self.height = self._raw.shape[0]
            self.width = self._raw.shape[1]
        return self._raw

    def set_data(self, data: np.ndarray):
        if self._raw.dtype != np.uint8:
            raise AssertionError(f"Expected format {np.uint8}. Received {data.dtype}.")

        self._raw = data
        self.height = self._raw.shape[0]
        self.width = self._raw.shape[1]

    def __repr__(self):
        res = f"{self.__module__}.{self.__class__.__name__}("
        res += f"\n  id={self.id}"
        res += f"\n  file_path={self.file_path}"
        res += "\n  data=" + (
            f"{self._raw.shape}, {self._raw.dtype}, min({np.min(self._raw)}), max({np.max(self._raw)})"
            if self._raw is not None
            else "<not-loaded>"
        )
        res += f"\n  width={self.width}"
        res += f"\n  height={self.height}"
        res += f"\n  description={self.description}"
        res += f"\n  ref_field={self.ref_field}"
        res += f"\n  key_name={self.key_name}"
        res += "\n)"
        return res


class Mask(Image):
    """Used for Semantic Segmentation. Same as Image class, but whenever rescaled, Image.Nearest strategy is used."""

    pass


class InstanceMask(Image, InstanceField):
    """
    Represents an instance mask that can be attached to any other field and can
    act as an input or a target.

    Args:
        file_path (str): The location of the image to be loaded.
        data (np.ndarray): The numpy array with the loaded image.
        width (Optional[int]): Width of the image.
        height (Optional[int]): Height of the image.
        instance_id (id): Id of the instance within the dataset
        label:( Optional[AnnotationClass]): An AnnotationClass associated with the instance
        description (str): A description of the object or annotation information.
        ref_field (List[object]): A list of objects the current BaseField is refering to.
        key_name (str): A string key, by which the field can be identified.
        confidence (float): Confidence level of the field.
        user_attributes (Dict[str, object]): Optional dictionary containing additional data regarding the field.

    Returns:
        :class: `efemarai.fields.data_fields.InstanceMask`: An InstanceMask object.
    """

    def __init__(
        self,
        instance_id: int = None,
        label: AnnotationClass = None,
        file_path: str = None,
        data: np.ndarray = None,
        width: int = None,
        height: int = None,
        confidence: float = None,
        id: ObjectId = None,
        description: str = None,
        ref_field: list = None,
        key_name: str = None,
        user_attributes: Dict[str, object] = None,
    ):
        """The MRO is >>> ef.InstanceMask.__mro__
        (<class 'efemarai.fields.data_fields.InstanceMask'>,
         <class 'efemarai.fields.data_fields.Image'>,
         <class 'efemarai.fields.annotation_fields.InstanceField'>,
         <class 'efemarai.fields.base_fields.BaseField'>,
         <class 'object'>)
        """
        Image.__init__(
            self,
            file_path=file_path,
            data=data,
            width=width,
            height=height,
            confidence=confidence,
            id=id,
            description=description,
            ref_field=ref_field,
            key_name=key_name,
            user_attributes=user_attributes,
        )
        InstanceField.__init__(
            self,
            instance_id=instance_id,
            label=label,
            confidence=confidence,
            id=id,
            description=description,
            ref_field=ref_field,
            key_name=key_name,
            user_attributes=user_attributes,
        )

    def __repr__(self):
        res = f"{self.__module__}.{self.__class__.__name__}("
        res += f"\n  id={self.id}"
        res += f"\n  file_path={self.file_path}"
        res += "\n  data=" + (
            f"{self._raw.shape}, {self._raw.dtype}, min({np.min(self._raw)}), max({np.max(self._raw)})"
            if self._raw is not None
            else "<not-loaded>"
        )
        res += f"\n  confidence={self.confidence}"
        res += f"\n  label={self.label}"
        res += f"\n  width={self.width}"
        res += f"\n  height={self.height}"
        res += f"\n  description={self.description}"
        res += f"\n  ref_field={self.ref_field}"
        res += f"\n  key_name={self.key_name}"
        res += "\n)"
        return res

    def to_polygon(self):
        """
        Converts `efemarai.fields.data_fields.InstanceMask` to `efemarai.fields.annotation_fields.Polygon`

        Returns:
            :class: `efemarai.fields.annotation_fields.Polygon`

        Example: See ef.Polygon for an example usage.
        """
        vertices, area = create_polygons_from_mask(self.data)
        return Polygon(
            vertices=vertices,
            area=area,
            instance_id=self.instance_id,
            label=self.label,
            confidence=self.confidence,
            id=self.id,
            description=self.description,
            ref_field=self.ref_field,
            key_name=self.key_name,
        )


class VideoFrame(Image):
    def __init__(
        self,
        index: int,
        file_path: str,
        width: int = None,
        height: int = None,
        data: np.ndarray = None,
        confidence: float = None,
        id: ObjectId = None,
        description: str = None,
        ref_field: list = None,
        key_name: str = None,
        user_attributes: Dict[str, object] = None,
    ):
        Image.__init__(
            self,
            file_path=file_path,
            data=data,
            width=width,
            height=height,
            confidence=confidence,
            id=id,
            description=description,
            ref_field=ref_field,
            key_name=key_name,
            user_attributes=user_attributes,
        )

        self.index = index
        self.width = width
        self.height = height
        self._raw = data


class Video(BaseField):
    def __init__(
        self,
        file_path: str,
        data: np.ndarray = None,
        frames: list = None,
        frame_count: int = None,
        fps: int = None,
        width: int = None,
        height: int = None,
        confidence: float = None,
        id: ObjectId = None,
        description: str = None,
        ref_field: list = None,
        key_name: str = None,
        user_attributes: Dict[str, object] = None,
    ):
        if frames is None:
            frames = []
        BaseField.__init__(
            self,
            confidence=confidence,
            id=id,
            description=description,
            ref_field=ref_field,
            key_name=key_name,
            user_attributes=user_attributes,
        )

        self.file_path = file_path
        self.frames = frames
        self.frame_count = frame_count
        self.fps = fps
        self.width = width
        self.height = height
        self._raw = data
        self._index = None

    def __iter__(self):
        for frame in self.frames:
            yield frame

    def __len__(self):
        return len(self)

    def __getitem__(self, i):
        if isinstance(i, slice):
            self._index = list(range(i.start, i.stop, i.step if i.step else 1))
        elif isinstance(i, (int, list, tuple)):
            self._index = i
        else:
            raise AttributeError(
                f"List indices must be int, list or slice. Got {type(i)}"
            )
        return self

    @property
    def data(self):
        if self._raw is None:
            self._raw = np.stack([frame.data for frame in self.frames])
        return self._raw

    def set_data(self, data):
        self._raw = data


class Tensor(BaseField):
    def __init__(
        self,
        data: np.ndarray,
        confidence: float = None,
        id: ObjectId = None,
        description: str = None,
        ref_field: list = None,
        key_name: str = None,
        user_attributes: Dict[str, object] = None,
    ):
        super().__init__(
            confidence=confidence,
            id=id,
            description=description,
            ref_field=ref_field,
            key_name=key_name,
            user_attributes=user_attributes,
        )

        self.data = data


class Text(BaseField):
    def __init__(
        self,
        text: str,
        confidence: float = None,
        id: ObjectId = None,
        description: str = None,
        ref_field: list = None,
        key_name: str = None,
        user_attributes: Dict[str, object] = None,
    ):
        super().__init__(
            confidence=confidence,
            id=id,
            description=description,
            ref_field=ref_field,
            key_name=key_name,
            user_attributes=user_attributes,
        )

        self.text = text


class DataFrame(BaseField):
    def __init__(
        self,
        data,
        target_column: str,
        file_path: str = None,
        confidence: float = None,
        id: ObjectId = None,
        description: str = None,
        ref_field: list = None,
        key_name: str = None,
        user_attributes: Dict[str, object] = None,
    ):
        super().__init__(
            confidence=confidence,
            id=id,
            description=description,
            ref_field=ref_field,
            key_name=key_name,
            user_attributes=user_attributes,
        )

        self.data = data
        self.file_path = file_path
        self.target_column = target_column

    def split_inputs_targets(self, n_rows=1):
        iterations = math.ceil(self.data.shape[0] / n_rows)
        inputs, targets = [], []
        target_column = self._check_column(self.data, self.target_column)
        for chunk in range(iterations):
            rows = self.data.loc[chunk * n_rows : ((chunk + 1) * n_rows - 1)]

            # exclude target column
            inputs.append(rows.loc[:, rows.columns != target_column])
            targets.append(rows[target_column].tolist())

        return inputs, targets

    def _serialize(self):
        self.data = self.data.to_dict("list")
        return sdk_serialize(self)

    @staticmethod
    def _check_column(data, target_column):
        if target_column not in data.columns:
            raise AttributeError(
                f"Target column '{target_column}' not found in dataframe columns: {data.columns.tolist()}."
                "Please provide a valid column name."
            )
        return target_column


class JSON(BaseField):
    def __init__(
        self,
        data: dict,
        confidence: float = None,
        id: ObjectId = None,
        description: str = None,
        ref_field: list = None,
        key_name: str = None,
        user_attributes: Dict[str, object] = None,
    ):
        super().__init__(
            confidence=confidence,
            id=id,
            description=description,
            ref_field=ref_field,
            key_name=key_name,
            user_attributes=user_attributes,
        )

        self.data = data


class Audio(BaseField):
    def __init__(
        self,
        data,
        confidence: float = None,
        id: ObjectId = None,
        description: str = None,
        ref_field: list = None,
        key_name: str = None,
        user_attributes: Dict[str, object] = None,
    ):
        super().__init__(
            confidence=confidence,
            id=id,
            description=description,
            ref_field=ref_field,
            key_name=key_name,
            user_attributes=user_attributes,
        )

        self.data = data



from efemarai.job_state import JobState
from efemarai.model import Model


class Baseline:
    """Provides Baseline related functionality."""

    @staticmethod
    def create(
        project,
        name,
        model,
        dataset,
    ):
        """Create a baseline. A more convenient way is to use :func:`project.create_baseline`"""
        if isinstance(model, str):
            model = project.model(model)

        if isinstance(dataset, str):
            dataset = project.dataset(dataset)

        response = project._put(
            "api/baselineRunById",
            json={
                "name": name,
                "model": model.id,
                "dataset": dataset.id,
            },
        )
        return Baseline(
            project,
            response["id"],
            name,
            model,
            dataset,
            "NotStarted",
            None,
            {},
        )

    def __init__(
        self, project, id, name, model, dataset, state, state_message, reports
    ):
        """Create a standard baseline evaluation."""
        self.project = (
            project  #: (:class:`efemarai.project.Project`) Associated project.
        )
        self.id = id
        self.name = name  #: (str) Name of the baseline.

        if isinstance(model, str):
            self._model = None
            self._model_id = model
        elif isinstance(model, dict):
            # TODO: Create a model.reload() to fetch model repository and files
            self._model = Model(
                **model, project=self.project, repository=None, files=None
            )
            self._model_id = self._model.id
        else:
            self._model = model
            self._model_id = self._model.id

        if isinstance(dataset, str):
            self._dataset = None
            self._dataset_id = dataset
        else:
            self._dataset = dataset
            self._dataset_id = dataset.id

        self.state = JobState(
            state
        )  # (:class:`efemarai.stress_test.JobState`) State of the baseline.

        self.state_message = (
            state_message  # (str) Optional message associated with the state
        )

        self._reports = reports

    def __repr__(self):
        res = f"{self.__module__}.{self.__class__.__name__}("
        res += f"\n  id={self.id}"
        res += f"\n  name={self.name}"
        res += f"\n  model={self.model.name}"
        res += f"\n  dataset={self.dataset.name}"
        res += f"\n  state={self.state}"
        res += f"\n  state_message={self.state_message}"
        res += f"\n  len(reports)={len(self._reports)}"
        res += "\n)"
        return res

    @property
    def model(self):
        """Returns the model associated with the baseline."""
        if self._model is None:
            self._model = next(m for m in self.project.models if m.id == self._model_id)
        return self._model

    @property
    def dataset(self):
        """Returns the dataset associated with the baseline."""
        if self._dataset is None:
            self._dataset = next(
                d for d in self.project.datasets if d.id == self._dataset_id
            )
        return self._dataset

    @property
    def finished(self):
        """Returns if the baseline has successfully finished.

        :rtype: bool
        """
        return self.state == JobState.Finished

    @property
    def failed(self):
        """Returns if the baseline has failed.

        :rtype: bool
        """
        return self.state == JobState.Failed

    @property
    def running(self):
        """Returns if the baseline is still running - not failed or finished.

        :rtype: bool
        """
        return self.state not in (JobState.Finished, JobState.Failed)

    def delete(self, delete_dependants=False):
        """Deletes a baseline run. This cannot be undone."""
        self.project._delete(f"api/baselineRunById/{self.id}/{delete_dependants}")

    def reload(self):
        """
        Reloads the baseline *in place* from the remote endpoint and return it.

        Rerturns:
            The updated baseline object.
        """
        response = self.project._get(f"api/baselineRunById/{self.id}")["value"]

        self.name = response["name"]
        self.state = JobState(response["states"][-1]["name"])
        self.state_message = response["states"][-1].get("message")
        self._reports = response.get("reports", {})

        return self
